@@ -16,10 +16,12 @@
  */
 /* TODO: Arc, Circle, Ellipsis, Bulge (Curve) tessellation.
  *       ocs/ucs transforms, explode of inserts?
- *       NOCOMMA and \n not with stdout. stdout is line-buffered (#75),
- *       so NOCOMMA cannot backup past the previous \n to delete the comma.
- *       We really have to add the comma before, not after, and special case
- * the first field, not the last to omit the comma.
+ *       NOCOMMA:
+ *         We really have to add the comma before, not after, and special case
+ *         the first field, not the last to omit the comma.
+ *       GeoJSON 2008 or newer RFC7946 https://tools.ietf.org/html/rfc7946#appendix-B
+ *       For the new format we need to follow the right-hand rule for orientation
+ *       (counterclockwise polygons).
  */
 
 #include "config.h"
@@ -30,6 +32,7 @@
 #include <math.h>
 #include <assert.h>
 
+#define IS_PRINT
 #include "common.h"
 #include "bits.h"
 #include "myalloca.h"
@@ -44,6 +47,11 @@
 /* the current version per spec block */
 static unsigned int cur_ver = 0;
 
+/* https://tools.ietf.org/html/rfc7946#section-11.2 */
+// The default %f is good enough
+//#undef FORMAT_RD
+//#define FORMAT_RD %.6f
+
 /*--------------------------------------------------------------------------------
  * See http://geojson.org/geojson-spec.html
  * Arc, AttributeDefinition, BlockReference, Ellipse, Hatch, Line,
@@ -62,8 +70,10 @@ static unsigned int cur_ver = 0;
            },
          "geometry":
            { "type": "LineString",
-             "coordinates": [ [ 370.858611653430728, 730.630303522043732, 0.0
- ], [ 450.039756420260289, 619.219273076899071, 0.0 ] ]
+             "coordinates": [
+               [ 370.858611, 730.630303 ],
+               [ 450.039756, 619.219273 ]
+             ]
            }
        },
      ], ...
@@ -73,7 +83,6 @@ static unsigned int cur_ver = 0;
  */
 
 #define ACTION geojson
-#define IS_PRINT
 
 #define PREFIX                                                                \
   for (int _i = 0; _i < dat->bit; _i++)                                       \
@@ -126,14 +135,16 @@ static unsigned int cur_ver = 0;
     dat->bit++;                                                               \
   }
 #define ENDSEC() ENDARRAY
-#define NOCOMMA fseek (dat->fh, -2, SEEK_CUR)
-#define PAIR_S(name, str)                                                     \
+#define OLD_NOCOMMA fseek (dat->fh, -2, SEEK_CUR)
+#define NOCOMMA assert(0 = "NOCOMMA")
+// guaranteed non-null str
+#define PAIR_Sc(name, str)                                                    \
   {                                                                           \
     const int len = strlen (str);                                             \
     if (len < 4096 / 6)                                                       \
       {                                                                       \
         const int _len = 6 * len + 1;                                         \
-        char *_buf = alloca (_len);                                           \
+        char *_buf = (char *)alloca (_len);                                   \
         PREFIX fprintf (dat->fh, "\"" #name "\": \"%s\",\n",                  \
                         json_cquote (_buf, str, _len));                       \
         freea (_buf);                                                         \
@@ -141,20 +152,28 @@ static unsigned int cur_ver = 0;
     else                                                                      \
       {                                                                       \
         const int _len = 6 * len + 1;                                         \
-        char *_buf = malloc (_len);                                           \
+        char *_buf = (char *)malloc (_len);                                   \
         PREFIX fprintf (dat->fh, "\"" #name "\": \"%s\",\n",                  \
                         json_cquote (_buf, str, _len));                       \
         free (_buf);                                                          \
       }                                                                       \
   }
+#define PAIR_S(name, str)                                                     \
+  if (str)                                                                    \
+    PAIR_Sc (name, str)
 #define PAIR_D(name, value)                                                   \
   {                                                                           \
     PREFIX fprintf (dat->fh, "\"" #name "\": %d,\n", value);                  \
   }
-#define LASTPAIR_S(name, value)                                               \
+// guaranteed non-null str
+#define LASTPAIR_Sc(name, value)                                              \
   {                                                                           \
     PREFIX fprintf (dat->fh, "\"" #name "\": \"%s\"\n", value);               \
   }
+#define LASTPAIR_S(name, value)                                               \
+  if (value) {                                                                \
+    PREFIX fprintf (dat->fh, "\"" #name "\": \"%s\"\n", value);               \
+  }
 #define PAIR_NULL(name)                                                       \
   {                                                                           \
     PREFIX fprintf (dat->fh, "\"" #name "\": null,\n");                       \
@@ -180,13 +199,8 @@ static unsigned int cur_ver = 0;
 //#define VALUE_RC(value,dxf) VALUE(value, RC, dxf)
 
 #define FIELD(name, type, dxf)
-//    PREFIX fprintf(dat->fh, "\"" #name "\": " FORMAT_##type ",\n",
-//    _obj->name)
 #define _FIELD(name, type, value)
-//    PREFIX fprintf(dat->fh, "\"" #name "\": " FORMAT_##type ",\n", obj->name)
 #define ENT_FIELD(name, type, value)
-//    PREFIX fprintf(dat->fh, "\"" #name "\": " FORMAT_##type ",\n",
-//    _ent->name)
 #define FIELD_CAST(name, type, cast, dxf) FIELD (name, cast, dxf)
 #define FIELD_TRACE(name, type)
 #define FIELD_TEXT(name, str)
@@ -204,10 +218,10 @@ static unsigned int cur_ver = 0;
 #define FIELD_BS(name, dxf) FIELD (name, BS, dxf)
 #define FIELD_BL(name, dxf) FIELD (name, BL, dxf)
 #define FIELD_BLL(name, dxf) FIELD (name, BLL, dxf)
-#define FIELD_BD(name, dxf) FIELD (name, BD, dxf)
+#define FIELD_BD(name, dxf)
 #define FIELD_RC(name, dxf) FIELD (name, RC, dxf)
 #define FIELD_RS(name, dxf) FIELD (name, RS, dxf)
-#define FIELD_RD(name, dxf) FIELD (name, RD, dxf)
+#define FIELD_RD(name, dxf) FIELD_BD (name, dxf)
 #define FIELD_RL(name, dxf) FIELD (name, RL, dxf)
 #define FIELD_RLL(name, dxf) FIELD (name, RLL, dxf)
 #define FIELD_MC(name, dxf) FIELD (name, MC, dxf)
@@ -219,12 +233,10 @@ static unsigned int cur_ver = 0;
 #define FIELD_T(name, dxf)
 //  { if (dat->version >= R_2007) { FIELD_TU(name, dxf); }
 //    else                        { FIELD_TV(name, dxf); } }
-#define FIELD_BT(name, dxf) FIELD (name, BT, dxf);
+#define FIELD_BT(name, dxf) FIELD (name, BT, dxf)
 #define FIELD_4BITS(name, dxf) FIELD (name, 4BITS, dxf)
 #define FIELD_BE(name, dxf) FIELD_3RD (name, dxf)
-#define FIELD_DD(name, _default, dxf)                                         \
-  PREFIX fprintf (dat->fh, FORMAT_DD ", ", _obj->name)
-#define FIELD_2DD(name, d1, d2, dxf)
+#define FIELD_2DD(name, def, dxf)
 #define FIELD_3DD(name, def, dxf)
 #define FIELD_2RD(name, dxf)
 #define FIELD_2BD(name, dxf)
@@ -232,24 +244,79 @@ static unsigned int cur_ver = 0;
 #define FIELD_3RD(name, dxf) ;
 #define FIELD_3BD(name, dxf)
 #define FIELD_3BD_1(name, dxf)
+#define FIELD_DD(name, _default, dxf)
+
+#define _VALUE_RD(value) fprintf (dat->fh, FORMAT_RD, value)
+#ifdef IS_RELEASE
+#  define VALUE_RD(value)                                                     \
+  {                                                                           \
+    if (bit_isnan (value))                                                    \
+      _VALUE_RD (0.0);                                                        \
+    else                                                                      \
+      _VALUE_RD (value);                                                      \
+   }
+#else
+#  define VALUE_RD(value) _VALUE_RD (value)
+#endif
 #define VALUE_2DPOINT(px, py)                                                 \
-  PREFIX fprintf (dat->fh, "[ " FORMAT_DD ", " FORMAT_DD " ],\n", px, py)
+  {                                                                           \
+    PREFIX fprintf (dat->fh, "[ ");                                           \
+    VALUE_RD (px);                                                            \
+    fprintf (dat->fh, ", ");                                                  \
+    VALUE_RD (py);                                                            \
+    fprintf (dat->fh, " ],\n");                                               \
+  }
 #define LASTVALUE_2DPOINT(px, py)                                             \
-  PREFIX fprintf (dat->fh, "[ " FORMAT_DD ", " FORMAT_DD " ]\n", px, py)
+  {                                                                           \
+    PREFIX fprintf (dat->fh, "[ ");                                           \
+    VALUE_RD (px);                                                            \
+    fprintf (dat->fh, ", ");                                                  \
+    VALUE_RD (py);                                                            \
+    fprintf (dat->fh, " ]\n");                                                \
+  }
 #define FIELD_2DPOINT(name) VALUE_2DPOINT (_obj->name.x, _obj->name.y)
 #define LASTFIELD_2DPOINT(name) LASTVALUE_2DPOINT (_obj->name.x, _obj->name.y)
-#define VALUE_3DPOINT(px, py, pz)                                             \
-  PREFIX fprintf (dat->fh,                                                    \
-                  "[ " FORMAT_DD ", " FORMAT_DD ", " FORMAT_DD " ],\n", px,   \
-                  py, pz)
+#define VALUE_3DPOINT(px, py, pz)                                          \
+  {                                                                           \
+    PREFIX fprintf (dat->fh, "[ ");                                           \
+    VALUE_RD (px);                                                            \
+    fprintf (dat->fh, ", ");                                                  \
+    VALUE_RD (py);                                                            \
+    if (pz != 0.0)                                                            \
+      {                                                                       \
+        fprintf (dat->fh, ", ");                                              \
+        VALUE_RD (pz);                                                        \
+      }                                                                       \
+    fprintf (dat->fh, " ],\n");                                               \
+  }
 #define LASTVALUE_3DPOINT(px, py, pz)                                         \
-  PREFIX fprintf (dat->fh,                                                    \
-                  "[ " FORMAT_DD ", " FORMAT_DD ", " FORMAT_DD " ]\n", px,    \
-                  py, pz)
+  {                                                                           \
+    PREFIX fprintf (dat->fh, "[ ");                                           \
+    VALUE_RD (px);                                                            \
+    fprintf (dat->fh, ", ");                                                  \
+    VALUE_RD (py);                                                            \
+    if (pz != 0.0)                                                            \
+      {                                                                       \
+        fprintf (dat->fh, ", ");                                              \
+        VALUE_RD (pz);                                                        \
+      }                                                                       \
+    fprintf (dat->fh, " ]\n");                                                \
+  }
 #define FIELD_3DPOINT(name)                                                   \
-  VALUE_3DPOINT (_obj->name.x, _obj->name.y, _obj->name.z)
+  {                                                                           \
+    if (_obj->name.z != 0.0)                                                  \
+      VALUE_3DPOINT (_obj->name.x, _obj->name.y, _obj->name.z)                \
+    else                                                                      \
+      FIELD_2DPOINT(name)                                                     \
+  }
 #define LASTFIELD_3DPOINT(name)                                               \
-  LASTVALUE_3DPOINT (_obj->name.x, _obj->name.y, _obj->name.z)
+  {                                                                           \
+    if (_obj->name.z != 0.0)                                                  \
+      LASTVALUE_3DPOINT (_obj->name.x, _obj->name.y, _obj->name.z)            \
+    else                                                                      \
+      LASTFIELD_2DPOINT (name)                                                \
+  }
+
 #define FIELD_CMC(name, dxf1, dxf2)
 #define FIELD_TIMEBLL(name, dxf)
 
@@ -260,20 +327,21 @@ static unsigned int cur_ver = 0;
   ARRAY;                                                                      \
   for (vcount = 0; vcount < (BITCODE_BL)size; vcount++)                       \
     {                                                                         \
-      PREFIX fprintf (dat->fh, "\"" #name "\": " FORMAT_##type ",\n",         \
-                      _obj->name[vcount]);                                    \
+      PREFIX fprintf (dat->fh, "\"" #name "\": " FORMAT_##type "%s\n",        \
+                      _obj->name[vcount],                                     \
+                      vcount == (BITCODE_BL)size - 1 ? "" : ",");             \
     }                                                                         \
-  if (size)                                                                   \
-    NOCOMMA;                                                                  \
   ENDARRAY;
-#define FIELD_VECTOR_T(name, size, dxf)                                       \
+
+#define FIELD_VECTOR_T(name, type, size, dxf)                                 \
   ARRAY;                                                                      \
-  PRE (R_2007)                                                                \
+  if (!(IS_FROM_TU (dat)))                                                    \
   {                                                                           \
     for (vcount = 0; vcount < (BITCODE_BL)_obj->size; vcount++)               \
       {                                                                       \
-        PREFIX fprintf (dat->fh, "\"" #name "\": \"%s\",\n",                  \
-                        _obj->name[vcount]);                                  \
+        PREFIX fprintf (dat->fh, "\"" #name "\": \"%s\"%s\n",                 \
+                        _obj->name[vcount],                                   \
+                        vcount == (BITCODE_BL)_obj->size - 1 ? "" : ",");     \
       }                                                                       \
   }                                                                           \
   else                                                                        \
@@ -281,43 +349,23 @@ static unsigned int cur_ver = 0;
     for (vcount = 0; vcount < (BITCODE_BL)_obj->size; vcount++)               \
       FIELD_TEXT_TU (name, _obj->name[vcount]);                               \
   }                                                                           \
-  if (_obj->size)                                                             \
-    NOCOMMA;                                                                  \
   ENDARRAY;
 
 #define FIELD_VECTOR(name, type, size, dxf)                                   \
   FIELD_VECTOR_N (name, type, _obj->size, dxf)
 
-#define FIELD_2RD_VECTOR(name, size, dxf)                                     \
-  ARRAY;                                                                      \
-  for (vcount = 0; vcount < (BITCODE_BL)_obj->size; vcount++)                 \
-    {                                                                         \
-      FIELD_2RD (name[vcount], dxf);                                          \
-    }                                                                         \
-  if (_obj->size)                                                             \
-    NOCOMMA;                                                                  \
-  ENDARRAY;
-
-#define FIELD_2DD_VECTOR(name, size, dxf)                                     \
-  ARRAY;                                                                      \
-  FIELD_2RD (name[0], 0);                                                     \
-  for (vcount = 1; vcount < (BITCODE_BL)_obj->size; vcount++)                 \
-    {                                                                         \
-      FIELD_2DD (name[vcount], FIELD_VALUE (name[vcount - 1].x),              \
-                 FIELD_VALUE (name[vcount - 1].y), dxf);                      \
-    }                                                                         \
-  if (_obj->size)                                                             \
-    NOCOMMA;                                                                  \
-  ENDARRAY;
+#define FIELD_2RD_VECTOR(name, size, dxf)
+#define FIELD_2DD_VECTOR(name, size, dxf)
 
 #define FIELD_3DPOINT_VECTOR(name, size, dxf)                                 \
   ARRAY;                                                                      \
   for (vcount = 0; vcount < (BITCODE_BL)_obj->size; vcount++)                 \
     {                                                                         \
-      FIELD_3DPOINT (name[vcount], dxf);                                      \
+      if (vcount == (BITCODE_BL)_obj->size - 1)                               \
+        LASTFIELD_3DPOINT (name[vcount], dxf)                                 \
+      else                                                                    \
+        FIELD_3DPOINT (name[vcount], dxf)                                     \
     }                                                                         \
-  if (_obj->size)                                                             \
-    NOCOMMA;                                                                  \
   ENDARRAY;
 
 #define WARN_UNSTABLE_CLASS                                                   \
@@ -326,6 +374,38 @@ static unsigned int cur_ver = 0;
             klass->proxyflag, klass->is_zombie ? "is_zombie" : "",          \
             obj->address + obj->size)
 
+// ensure counter-clockwise orientation of a closed polygon. 2d only.
+static int
+normalize_polygon_orient (BITCODE_BL numpts, dwg_point_2d **const pts_p)
+{
+  double sum = 0.0;
+  dwg_point_2d *pts = *pts_p;
+  // check orientation
+  for (unsigned i = 0; i < numpts - 1; i++)
+    {
+      sum += (pts[i+1].x - pts[i].x) * (pts[i+1].y + pts[i].y);
+    }
+  if (sum > 0.0) // if clockwise
+    {
+      // reverse and return a copy
+      unsigned last = numpts - 1;
+      dwg_point_2d *new = malloc (numpts * sizeof (BITCODE_2RD));
+      //fprintf (stderr, "%u pts, sum %f: reverse orient\n", numpts, sum);
+      for (unsigned i = 0; i < numpts; i++)
+        {
+          new[i].x = pts[last-i].x;
+          new[i].y = pts[last-i].y;
+        }
+      *pts_p = new;
+      return 1;
+    }
+  else
+    {
+      //fprintf (stderr, "%u pts, sum %f: keep orient\n", numpts, sum);
+      return 0;
+    }
+}
+
 // common properties
 static void
 dwg_geojson_feature (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
@@ -335,7 +415,9 @@ dwg_geojson_feature (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
   char *name;
   char tmp[64];
 
-  PAIR_S (type, "Feature");
+  PAIR_Sc (type, "Feature");
+  sprintf (tmp, "%lX", obj->handle.value);
+  PAIR_Sc (id, tmp);
   KEY (properties);
   SAMEHASH;
   PAIR_S (SubClasses, subclass);
@@ -351,48 +433,83 @@ dwg_geojson_feature (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
           if (!error)
             {
               PAIR_S (Layer, name);
-              if (dat->version >= R_2007)
+              if (IS_FROM_TU (dat))
                 free (name);
             }
         }
 
       // See #95: index as int or rgb as hexstring
-      if (dat->version >= R_2004)
+      if (dat->version >= R_2004
+          && (obj->tio.entity->color.method == 0xc3     // Truecolor
+              || obj->tio.entity->color.method == 0xc2) // Entity
+          && obj->tio.entity->color.index == 256)
         {
-          sprintf (tmp, "%06X", obj->tio.entity->color.rgb & 0xffffff);
-          PAIR_S (color, tmp);
+          sprintf (tmp, "#%06X", obj->tio.entity->color.rgb & 0xffffff);
+          PAIR_Sc (Color, tmp);
         }
-      else
+      else if ((obj->tio.entity->color.index != 256)
+               || (dat->version >= R_2004
+                   && obj->tio.entity->color.method != 0xc0 // ByLayer
+                   && obj->tio.entity->color.method != 0xc1 // ByBlock
+                   && obj->tio.entity->color.method != 0xc8 // none
+                   ))
         {
-          PAIR_D (color, obj->tio.entity->color.index);
+          // no names for the first palette entries yet.
+          PAIR_D (Color, obj->tio.entity->color.index);
         }
 
       name = dwg_ent_get_ltype_name (obj->tio.entity, &error);
       if (!error && strNE (name, "ByLayer")) // skip the default
         {
           PAIR_S (Linetype, name);
-          if (dat->version >= R_2007)
+          if (IS_FROM_TU (dat))
             free (name);
         }
     }
 
-  sprintf (tmp, "%lX", obj->handle.value);
-  // if has name or text
+  // if has notes and opt. an mtext frame_text
   if (obj->type == DWG_TYPE_GEOPOSITIONMARKER)
     {
       Dwg_Entity_GEOPOSITIONMARKER *_obj
           = obj->tio.entity->tio.GEOPOSITIONMARKER;
-      PAIR_S (Text, _obj->text)
+      if (IS_FROM_TU (dat))
+        {
+          char *utf8 = bit_convert_TU ((BITCODE_TU)_obj->notes);
+          PAIR_S (Text, utf8)
+          free (utf8);
+        }
+      else
+        {
+          PAIR_S (Text, _obj->notes)
+        }
     }
   else if (obj->type == DWG_TYPE_TEXT)
     {
       Dwg_Entity_TEXT *_obj = obj->tio.entity->tio.TEXT;
-      PAIR_S (Text, _obj->text_value)
+      if (IS_FROM_TU (dat))
+        {
+          char *utf8 = bit_convert_TU ((BITCODE_TU)_obj->text_value);
+          PAIR_S (Text, utf8)
+          free (utf8);
+        } 
+      else
+        {      
+          PAIR_S (Text, _obj->text_value)
+        }
     }
   else if (obj->type == DWG_TYPE_MTEXT)
     {
       Dwg_Entity_MTEXT *_obj = obj->tio.entity->tio.MTEXT;
-      PAIR_S (Text, _obj->text)
+      if (IS_FROM_TU (dat))
+        {
+          char *utf8 = bit_convert_TU ((BITCODE_TU)_obj->text);
+          PAIR_S (Text, utf8)
+          free (utf8);
+        } 
+      else
+        {      
+          PAIR_S (Text, _obj->text)
+        }
     }
   else if (obj->type == DWG_TYPE_INSERT)
     {
@@ -402,14 +519,14 @@ dwg_geojson_feature (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
         {
           Dwg_Object_BLOCK_HEADER *_hdr = hdr->tio.object->tio.BLOCK_HEADER;
           char *text;
-          if (dat->version >= R_2007)
+          if (IS_FROM_TU (dat))
             text = bit_convert_TU ((BITCODE_TU)_hdr->name);
           else
             text = _hdr->name;
           if (text)
             {
               PAIR_S (name, text);
-              if (dat->version >= R_2007)
+              if (IS_FROM_TU (dat))
                 free (text);
             }
         }
@@ -422,55 +539,80 @@ dwg_geojson_feature (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
         {
           Dwg_Object_BLOCK_HEADER *_hdr = hdr->tio.object->tio.BLOCK_HEADER;
           char *text;
-          if (dat->version >= R_2007)
+          if (IS_FROM_TU (dat))
             text = bit_convert_TU ((BITCODE_TU)_hdr->name);
           else
             text = _hdr->name;
           if (text)
             {
               PAIR_S (name, text);
-              if (dat->version >= R_2007)
+              if (IS_FROM_TU (dat))
                 free (text);
             }
         }
     }
   // PAIR_NULL(ExtendedEntity);
-  LASTPAIR_S (EntityHandle, tmp);
+  sprintf (tmp, "%lX", obj->handle.value);
+  LASTPAIR_Sc (EntityHandle, tmp);
   ENDHASH;
 }
 
 #define FEATURE(subclass, obj)                                                \
   HASH;                                                                       \
   dwg_geojson_feature (dat, obj, #subclass)
-#define ENDFEATURE ENDHASH
+#define ENDFEATURE                                                            \
+  if (is_last)                                                                \
+    LASTENDHASH                                                               \
+  else                                                                        \
+    ENDHASH
 
 static int
-dwg_geojson_LWPOLYLINE (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
+dwg_geojson_LWPOLYLINE (Bit_Chain *restrict dat, Dwg_Object *restrict obj, int is_last)
 {
   BITCODE_BL j, last_j;
   Dwg_Entity_LWPOLYLINE *_obj = obj->tio.entity->tio.LWPOLYLINE;
-  last_j = _obj->num_points - 1;
-  // TODO: if closed and num_points > 3 use a Polygon
+  dwg_point_2d *pts = (dwg_point_2d*)_obj->points;
+  if (!_obj->points)
+    return 1;
+
   FEATURE (AcDbEntity : AcDbLwPolyline, obj);
-  GEOMETRY (LineString);
-  KEY (coordinates);
-  ARRAY;
-  for (j = 0; j < last_j; j++)
+  // if closed and num_points > 3 use a Polygon
+  if (_obj->flag & 512 && _obj->num_points > 3)
     {
-      FIELD_2DPOINT (points[j]);
+      int changed = normalize_polygon_orient (_obj->num_points, &pts); // RFC7946
+      GEOMETRY (Polygon)
+      KEY (coordinates);
+      ARRAY;
+      ARRAY;
+      for (j = 0; j < _obj->num_points; j++)
+        VALUE_2DPOINT (pts[j].x, pts[j].y)
+      LASTVALUE_2DPOINT (pts[0].x, pts[0].y);
+      LASTENDARRAY;
+      LASTENDARRAY;
+      if (changed)
+        free (pts);
+    }
+  else
+    {
+      GEOMETRY (LineString)
+      KEY (coordinates);
+      ARRAY;
+      last_j = _obj->num_points - 1;
+      for (j = 0; j < last_j; j++)
+        VALUE_2DPOINT (pts[j].x, pts[j].y);
+      LASTVALUE_2DPOINT (pts[last_j].x, pts[last_j].y);
+      LASTENDARRAY;
     }
-  LASTFIELD_2DPOINT (points[last_j]);
-  LASTENDARRAY;
   ENDGEOMETRY;
   ENDFEATURE;
-  return 0;
+  return 1;
 }
 
 /* returns 0 if object could be printed
  */
 static int
 dwg_geojson_variable_type (Dwg_Data *restrict dwg, Bit_Chain *restrict dat,
-                           Dwg_Object *restrict obj)
+                           Dwg_Object *restrict obj, int is_last)
 {
   int i;
   char *dxfname;
@@ -479,7 +621,7 @@ dwg_geojson_variable_type (Dwg_Data *restrict dwg, Bit_Chain *restrict dat,
 
   i = obj->type - 500;
   if (i < 0 || i >= (int)dwg->num_classes)
-    return DWG_ERR_INVALIDTYPE;
+    return 0;
 
   klass = &dwg->dwg_class[i];
   if (!klass || !klass->dxfname)
@@ -490,17 +632,20 @@ dwg_geojson_variable_type (Dwg_Data *restrict dwg, Bit_Chain *restrict dat,
 
   if (strEQc (dxfname, "LWPOLYLINE"))
     {
-      return dwg_geojson_LWPOLYLINE (dat, obj);
+      return dwg_geojson_LWPOLYLINE (dat, obj, is_last);
     }
+  /*
   if (strEQc (dxfname, "GEODATA"))
     {
       Dwg_Object_GEODATA *_obj = obj->tio.object->tio.GEODATA;
       WARN_UNSTABLE_CLASS;
       FEATURE (AcDbObject : AcDbGeoData, obj);
-      // which fields?
+      // which fields? transformation for the world-coordinates?
+      // crs links of type proj4, ogcwkt, esriwkt or such?
       ENDFEATURE;
       return 0;
     }
+  */
   if (strEQc (dxfname, "GEOPOSITIONMARKER"))
     {
       Dwg_Entity_GEOPOSITIONMARKER *_obj
@@ -511,23 +656,19 @@ dwg_geojson_variable_type (Dwg_Data *restrict dwg, Bit_Chain *restrict dat,
       GEOMETRY (Point);
       KEY (coordinates);
       if (fabs (_obj->position.z) > 0.000001)
-        {
-          VALUE_3DPOINT (_obj->position.x, _obj->position.y, _obj->position.z);
-        }
+        VALUE_3DPOINT (_obj->position.x, _obj->position.y, _obj->position.z)
       else
-        {
-          VALUE_2DPOINT (_obj->position.x, _obj->position.y);
-        }
+        VALUE_2DPOINT (_obj->position.x, _obj->position.y);
       ENDGEOMETRY;
       ENDFEATURE;
-      return 0;
+      return 1;
     }
 
-  return DWG_ERR_UNHANDLEDCLASS;
+  return 0;
 }
 
-static void
-dwg_geojson_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
+static int
+dwg_geojson_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj, int is_last)
 {
   switch (obj->type)
     {
@@ -538,13 +679,11 @@ dwg_geojson_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
         // TODO: explode insert into a GeometryCollection
         GEOMETRY (Point);
         KEY (coordinates);
-        ARRAY;
         LASTFIELD_3DPOINT (ins_pt);
-        LASTENDARRAY;
         ENDGEOMETRY;
         ENDFEATURE;
+        return 1;
       }
-      break;
     case DWG_TYPE_MINSERT:
       // a grid of INSERT's (named points)
       // dwg_geojson_MINSERT(dat, obj);
@@ -554,41 +693,70 @@ dwg_geojson_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
       {
         int error;
         BITCODE_BL j, numpts;
-        dwg_point_2d *pts;
-        FEATURE (AcDbEntity : AcDbPolyline, obj);
-        GEOMETRY (MultiLineString);
-        KEY (coordinates);
-        ARRAY;
+        bool is_polygon = false;
+        int changed = 0;
+        dwg_point_2d *pts, *orig;
+        Dwg_Entity_POLYLINE_2D *_obj = obj->tio.entity->tio.POLYLINE_2D;
         numpts = dwg_object_polyline_2d_get_numpoints (obj, &error);
+        if (error || !numpts)
+          return 0;
         pts = dwg_object_polyline_2d_get_points (obj, &error);
-        for (j = 0; j < numpts; j++)
+        if (error || !pts)
+          return 0;
+        // if closed and num_points > 3 use a Polygon
+        FEATURE (AcDbEntity : AcDbPolyline, obj);
+        if (_obj->flag & 512 && numpts > 3)
           {
-            if (j == numpts - 1)
+            orig = pts; // pts is already a new copy
+            changed = normalize_polygon_orient (numpts, &pts); // RFC7946
+            if (changed)
+              free (orig);
+            GEOMETRY (Polygon)
+            KEY (coordinates);
+            ARRAY;
+            ARRAY;
+            for (j = 0; j < numpts; j++)
+              VALUE_2DPOINT (pts[j].x, pts[j].y)
+            LASTVALUE_2DPOINT (pts[0].x, pts[0].y);
+            LASTENDARRAY;
+            LASTENDARRAY;
+            if (changed)
+              free (pts);
+          }
+        else
+          {
+            GEOMETRY (LineString)
+            KEY (coordinates);
+            ARRAY;
+            for (j = 0; j < numpts; j++)
               {
-                LASTVALUE_2DPOINT (pts[j].x, pts[j].y);
+                if (j == numpts - 1)
+                  LASTVALUE_2DPOINT (pts[j].x, pts[j].y)
+                else
+                  VALUE_2DPOINT (pts[j].x, pts[j].y);
               }
-            else
-              {
-                VALUE_2DPOINT (pts[j].x, pts[j].y);
-              }
+            free (pts);
+            LASTENDARRAY;
           }
-        free (pts);
-        LASTENDARRAY;
         ENDGEOMETRY;
         ENDFEATURE;
+        return 1;
       }
-      break;
     case DWG_TYPE_POLYLINE_3D:
       {
         int error;
         BITCODE_BL j, numpts;
         dwg_point_3d *pts;
+        numpts = dwg_object_polyline_3d_get_numpoints (obj, &error);
+        if (error || !numpts)
+          return 0;
+        pts = dwg_object_polyline_3d_get_points (obj, &error);
+        if (error || !pts)
+          return 0;
         FEATURE (AcDbEntity : AcDbPolyline, obj);
-        GEOMETRY (MultiLineString);
+        GEOMETRY (LineString);
         KEY (coordinates);
         ARRAY;
-        numpts = dwg_object_polyline_3d_get_numpoints (obj, &error);
-        pts = dwg_object_polyline_3d_get_points (obj, &error);
         for (j = 0; j < numpts; j++)
           {
             if (j == numpts - 1)
@@ -604,8 +772,8 @@ dwg_geojson_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
         LASTENDARRAY;
         ENDGEOMETRY;
         ENDFEATURE;
+        return 1;
       }
-      break;
     case DWG_TYPE_ARC:
       // needs explosion of arc into lines
       // dwg_geojson_ARC(dat, obj);
@@ -628,8 +796,8 @@ dwg_geojson_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
         LASTENDARRAY;
         ENDGEOMETRY;
         ENDFEATURE;
+        return 1;
       }
-      break;
     case DWG_TYPE_POINT:
       {
         Dwg_Entity_POINT *_obj = obj->tio.entity->tio.POINT;
@@ -646,9 +814,10 @@ dwg_geojson_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
           }
         ENDGEOMETRY;
         ENDFEATURE;
+        return 1;
       }
-      break;
     case DWG_TYPE__3DFACE:
+      // really a Polygon
       // dwg_geojson__3DFACE(dat, obj);
       LOG_TRACE ("3DFACE not yet supported")
       break;
@@ -681,7 +850,7 @@ dwg_geojson_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
       break;
     case DWG_TYPE__3DSOLID:
       // dwg_geojson__3DSOLID(dat, obj);
-      break; /* Check the type of the object */
+      break;
     case DWG_TYPE_REGION:
       // dwg_geojson_REGION(dat, obj);
       break;
@@ -703,13 +872,11 @@ dwg_geojson_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
         FEATURE (AcDbEntity : AcDbText, obj);
         GEOMETRY (Point);
         KEY (coordinates);
-        ARRAY;
-        LASTFIELD_2DPOINT (insertion_pt);
-        LASTENDARRAY;
+        LASTFIELD_2DPOINT (ins_pt);
         ENDGEOMETRY;
         ENDFEATURE;
+        return 1;
       }
-      break;
     case DWG_TYPE_MTEXT:
       {
         // add Text property to a point
@@ -717,39 +884,45 @@ dwg_geojson_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
         FEATURE (AcDbEntity : AcDbMText, obj);
         GEOMETRY (Point);
         KEY (coordinates);
-        ARRAY;
-        LASTFIELD_3DPOINT (insertion_pt);
-        LASTENDARRAY;
+        LASTFIELD_3DPOINT (ins_pt);
         ENDGEOMETRY;
         ENDFEATURE;
+        return 1;
       }
-      break;
     case DWG_TYPE_MLINE:
       // dwg_geojson_MLINE(dat, obj);
       LOG_TRACE ("MLINE not yet supported")
       break;
     case DWG_TYPE_LWPOLYLINE:
-      (void)dwg_geojson_LWPOLYLINE (dat, obj);
-      break;
+      return dwg_geojson_LWPOLYLINE (dat, obj, is_last);
     default:
       if (obj->type != obj->parent->layout_type)
-        dwg_geojson_variable_type (obj->parent, dat, obj);
+        return dwg_geojson_variable_type (obj->parent, dat, obj, is_last);
     }
+  return 0;
 }
 
 static int
 geojson_entities_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 {
   BITCODE_BL i;
-
+  int success;
   SECTION (features);
   for (i = 0; i < dwg->num_objects; i++)
     {
+      int is_last = i == dwg->num_objects - 1;
       Dwg_Object *obj = &dwg->object[i];
-      dwg_geojson_object (dat, obj);
+      success = dwg_geojson_object (dat, obj, is_last);
+      if (is_last && !success) // needed for the LASTFEATURE comma. end with an empty dummy
+        {
+          HASH
+          PAIR_Sc (type, "Feature");
+          PAIR_NULL (properties);
+          LASTPAIR_NULL (geometry);
+          LASTENDHASH;
+        }
     }
-  NOCOMMA;
-  ENDSEC ();
+  ENDSEC (); // because afterwards is always the final geocoding object
   return 0;
 }
 
@@ -760,8 +933,11 @@ dwg_write_geojson (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
   char date[12] = "YYYY-MM-DD";
   time_t rawtime;
 
+  if (!dwg->num_objects)
+    goto fail;
+
   HASH;
-  PAIR_S (type, "FeatureCollection");
+  PAIR_Sc (type, "FeatureCollection");
 
   // array of features
   if (geojson_entities_write (dat, dwg))
@@ -771,15 +947,15 @@ dwg_write_geojson (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
   HASH;
   time (&rawtime);
   strftime (date, 12, "%Y-%m-%d", localtime (&rawtime));
-  PAIR_S (creation_date, date);
+  PAIR_Sc (creation_date, date);
   KEY (generator);
   HASH;
   KEY (author);
   HASH;
-  LASTPAIR_S (name, "dwgread");
+  LASTPAIR_Sc (name, "dwgread");
   ENDHASH;
-  PAIR_S (package, PACKAGE_NAME);
-  LASTPAIR_S (version, PACKAGE_VERSION);
+  PAIR_Sc (package, PACKAGE_NAME);
+  LASTPAIR_Sc (version, PACKAGE_VERSION);
   LASTENDHASH;
   // PAIR_S(license, "?");
   LASTENDHASH;
